class Error(Exception):
    """Base class for exceptions for this module.
    It is used to recognize errors specific to this module"""


class EWFError(Error):
    """Related to EWF (Expert Witness disk image Format)"""


class InvalidSnapshot(Error):
    """Related to ASDF (Acquire Snapshot Data Format)"""


class InvalidBlock(Error):
    """Related to ASDF (Acquire Snapshot Data Format)"""


class UnsupportedVersion(Error):
    """Related to ASDF (Acquire Snapshot Data Format)"""



from __future__ import annotations

import argparse
import hashlib
import io
import sys
import traceback
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO
from zlib import crc32

from dissect.util.stream import RangeStream

from dissect.evidence.asdf import asdf

if TYPE_CHECKING:
    from collections.abc import Iterator


def iter_fileobj(src: BinaryIO) -> Iterator[bytes]:
    buf = src.read(io.DEFAULT_BUFFER_SIZE)
    while buf:
        yield buf
        buf = src.read(io.DEFAULT_BUFFER_SIZE)


def hash_fileobj(src: BinaryIO, alg: str = "sha256") -> bytes:
    ctx = hashlib.new(alg)
    for buf in iter_fileobj(src):
        ctx.update(buf)
    return ctx.digest()


def crc32_filobj(src: BinaryIO) -> int:
    crc = 0
    for buf in iter_fileobj(src):
        crc = crc32(buf, crc) & 0xFFFFFFFF
    return crc


@contextmanager
def status(line: str, verbose: bool = False) -> Iterator:
    TEMPLATE = "[{}] {:<50}{}"
    try:
        print(TEMPLATE.format("*", line, ""), end="\n")
        sys.stdout.flush()
        yield
    except Exception as e:
        if verbose:
            traceback.print_exc()
        print(f"[!] {e}")


def main() -> int:
    parser = argparse.ArgumentParser(description="Utility to verify ASDF files")
    parser.add_argument("file", metavar="ASDF", help="ASDF file to verify")
    parser.add_argument("--skip-hash", action="store_true", help="skip file hash")
    parser.add_argument("--skip-blocks", action="store_true", help="skip block checking")
    parser.add_argument("-v", "--verbose", action="store_true", help="increase verbosity")
    args = parser.parse_args()

    with Path(args.file).open("rb") as fh:
        header = None
        footer = None
        footer_offset = 0

        with status("Checking header", args.verbose):
            header = asdf.c_asdf.header(fh)
            if header.magic != asdf.FILE_MAGIC:
                print("[!] Invalid header magic")
                return 1

        with status("Checking footer", args.verbose):
            fh.seek(-len(asdf.c_asdf.footer), io.SEEK_END)
            footer_offset = fh.tell()
            footer = asdf.c_asdf.footer(fh)
            if footer.magic != asdf.FOOTER_MAGIC:
                footer = None
                print("[!] Invalid footer magic, please run asdf-repair")
                return 1

        if not args.skip_hash and footer:
            with status("Checking file hash", args.verbose):
                hashstream = RangeStream(fh, 0, footer_offset)
                res = hash_fileobj(hashstream)
                if res != footer.sha256:
                    print(f"[!] File hash doesn't match. Expected {footer.sha256.hex()}, got {res.hex()}")
                    return 1
        else:
            print("[@] Skipping file hash")

        if not args.skip_blocks and footer:
            with status("Checking blocks", args.verbose):
                table_size = (footer_offset - footer.table_offset) // len(asdf.c_asdf.table_entry)
                fh.seek(footer.table_offset)
                table = asdf.c_asdf.table_entry[table_size](fh)
                for i, entry in enumerate(table):
                    if not entry.flags & asdf.c_asdf.BLOCK_FLAG.CRC32:
                        continue

                    fh.seek(entry.file_offset)
                    block_header = asdf.c_asdf.block(fh)
                    if block_header.magic != asdf.BLOCK_MAGIC:
                        print(f"[!] Block {i} has invalid block magic")

                    data_offset = fh.tell()
                    fh.seek(entry.file_size - 4, io.SEEK_CUR)
                    target_crc = asdf.c_asdf.uint32(fh)

                    block_fh = RangeStream(fh, data_offset, entry.file_size - 4)
                    crc = crc32_filobj(block_fh)
                    if crc != target_crc:
                        print(f"[!] Block {i} crc32 doesn't match. Expected 0x{target_crc:x}, got 0x{crc:x}")

        else:
            print("[@] Skipping block check")

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest

from dissect.evidence.asdf import AsdfWriter

if TYPE_CHECKING:
    from collections.abc import Iterator


def open_data(name: str) -> Iterator[BinaryIO]:
    with (Path(__file__).parent / name).open("rb") as fh:
        yield fh


@pytest.fixture
def ad1_data() -> Iterator[BinaryIO]:
    yield from open_data("data/ad1_test.ad1")


@pytest.fixture
def ad1_data_long() -> Iterator[BinaryIO]:
    yield from open_data("data/ad1_long.ad1")


@pytest.fixture
def ad1_data_compressed() -> Iterator[BinaryIO]:
    yield from open_data("data/ad1_test_compressed.ad1")


@pytest.fixture
def ewf_data() -> Iterator[BinaryIO]:
    yield from open_data("data/ewf.E01")


@pytest.fixture
def asdf_writer() -> AsdfWriter:
    def noop() -> None:
        pass

    fh = BytesIO()
    fh.close = noop  # Prevent clearing the buffer, we need it
    return AsdfWriter(fh)



from __future__ import annotations

import zlib
from io import BytesIO

import pytest

from dissect.evidence.asdf.asdf import AsdfSnapshot, AsdfWriter
from dissect.evidence.asdf.streams import CompressedStream, Crc32Stream, HashedStream
from dissect.evidence.exceptions import InvalidSnapshot


def test_asdf(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(b"\x00" * 0x1000, idx=0, base=0)
    asdf_writer.add_bytes(b"\x02" * 0x1000, idx=0, base=0x4000)
    asdf_writer.add_bytes(b"\x04" * 0x1000, idx=0, base=0x8000)
    asdf_writer.add_bytes(b"\x06" * 0x1000, idx=0, base=0x10000)
    asdf_writer.add_bytes(b"\xff" * 0x1000, idx=0, base=0x14000)

    asdf_writer.add_bytes(b"\x08" * 0x1000, idx=1, base=0x2000)
    asdf_writer.add_bytes(b"\x10" * 0x1000, idx=1, base=0x5000)
    asdf_writer.add_bytes(b"\x12" * 0x1000, idx=1, base=0x8000)
    asdf_writer.add_bytes(b"\x14" * 0x1000, idx=1, base=0xB000)
    asdf_writer.add_bytes(b"\xff" * 0x1000, idx=1, base=0xE000)

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)
    stream_0 = reader.open(0)
    assert [(run_start, run_size) for run_start, run_size, _, _ in stream_0.table] == [
        (0, 0x1000),
        (0x4000, 0x1000),
        (0x8000, 0x1000),
        (0x10000, 0x1000),
        (0x14000, 0x1000),
    ]

    assert stream_0.read(0x1000) == b"\x00" * 0x1000
    assert stream_0.read(0x1000) == (b"\xa5\xdf" * (0x1000 // 2))
    assert stream_0.read(0x4000) == (b"\xa5\xdf" * (0x2000 // 2)) + (b"\x02" * 0x1000) + (b"\xa5\xdf" * (0x1000 // 2))

    stream_0.seek(0)
    assert stream_0.read() == b"".join(
        [
            (b"\x00" * 0x1000),
            (b"\xa5\xdf" * (0x3000 // 2)),
            (b"\x02" * 0x1000),
            (b"\xa5\xdf" * (0x3000 // 2)),
            (b"\x04" * 0x1000),
            (b"\xa5\xdf" * (0x7000 // 2)),
            (b"\x06" * 0x1000),
            (b"\xa5\xdf" * (0x3000 // 2)),
            (b"\xff" * 0x1000),
        ]
    )

    stream_1 = reader.open(1)
    assert stream_1.read(0x4000) == (b"\xa5\xdf" * (0x2000 // 2)) + (b"\x08" * 0x1000) + (b"\xa5\xdf" * (0x1000 // 2))


def test_asdf_overlap(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(b"\x01" * 100, base=0)
    asdf_writer.add_bytes(b"\x02" * 100, base=200)
    assert asdf_writer._table_lookup[0] == [0, 200]

    asdf_writer.add_bytes(b"\x03" * 100, base=50)
    assert asdf_writer._table_lookup[0] == [0, 100, 200]

    asdf_writer.add_bytes(b"\x04" * 150, base=100)
    assert asdf_writer._table_lookup[0] == [0, 100, 150, 200]

    asdf_writer.add_bytes(b"\x05" * 50, base=25)
    assert asdf_writer._table_lookup[0] == [0, 100, 150, 200]

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)
    stream = reader.open(0)

    assert [(run_start, run_size) for run_start, run_size, _, _ in stream.table] == [
        (0, 100),
        (100, 50),
        (150, 50),
        (200, 100),
    ]
    assert stream.read() == (b"\x01" * 100) + (b"\x03" * 50) + (b"\x04" * 50) + (b"\x02" * 100)


def test_asdf_overlap_all(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(b"\x01" * 100, base=0)
    asdf_writer.add_bytes(b"\x02" * 100, base=200)
    asdf_writer.add_bytes(b"\x03" * 100, base=50)
    asdf_writer.add_bytes(b"\x04" * 150, base=100)
    assert asdf_writer._table_lookup[0] == [0, 100, 150, 200]
    asdf_writer.add_bytes(b"\x06" * 400, base=0)
    assert asdf_writer._table_lookup[0] == [0, 100]

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)
    stream = reader.open(0)

    assert [(run_start, run_size) for run_start, run_size, _, _ in stream.table] == [
        (0, 100),
        (100, 300),
    ]
    assert stream.read() == (b"\x01" * 100) + (b"\x06" * 300)


def test_asdf_overlap_contiguous(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(b"\x01" * 100, base=0)
    asdf_writer.add_bytes(b"\x02" * 100, base=100)
    assert asdf_writer._table_lookup[0] == [0, 100]

    asdf_writer.add_bytes(b"\x03" * 75, base=50)
    assert asdf_writer._table_lookup[0] == [0, 100]

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)
    stream = reader.open(0)

    assert [(run_start, run_size) for run_start, run_size, _, _ in stream.table] == [
        (0, 100),
        (100, 100),
    ]
    assert stream.read() == (b"\x01" * 100) + (b"\x02" * 100)


def test_asdf_overlap_seek(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(b"\x00" * 100, base=0)
    asdf_writer.add_bytes(b"\x00" * 100, base=200)
    asdf_writer.add_bytes(bytes(range(200)), base=50)
    assert asdf_writer._table_lookup[0] == [0, 100, 200]

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)
    stream = reader.open(0)

    assert [(run_start, run_size) for run_start, run_size, _, _ in stream.table] == [
        (0, 100),
        (100, 100),
        (200, 100),
    ]
    assert stream.read() == (b"\x00" * 100) + bytes(range(50, 150)) + (b"\x00" * 100)


def test_asdf_mid_run(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(bytes([v & 0xFF for v in range(4096)]), base=0)

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)
    stream = reader.open(0)
    stream.align = 512

    stream.seek(1100)
    assert stream.read(512) == bytes([v & 0xFF for v in range(1100, 1100 + 512)])


def test_asdf_metadata(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_metadata_file("file", BytesIO(b"content"))
    asdf_writer.add_metadata_file("dir/file", BytesIO(b"content here too"))

    asdf_writer.close()
    asdf_writer._fh.seek(0)

    reader = AsdfSnapshot(asdf_writer._fh)

    assert reader.metadata.names() == ["file", "dir/file"]
    assert reader.metadata.open("file").read() == b"content"
    assert reader.metadata.open("dir/file").read() == b"content here too"

    with pytest.raises(KeyError):
        reader.metadata.open("nonexistent")


def test_asdf_stream_crc32() -> None:
    fh = BytesIO()
    stream = Crc32Stream(fh)
    stream.write(b"srt was here")
    stream.finalize()

    assert fh.getvalue() == b"srt was here\x2f\x0e\x60\xa4"


def test_asdf_stream_compressed() -> None:
    fh = BytesIO()
    stream = CompressedStream(fh)
    stream.write(b"srt was here" * 100)
    stream.finalize()

    assert zlib.decompress(fh.getvalue()) == b"srt was here" * 100


def test_asdf_stream_hashed() -> None:
    fh = BytesIO()
    stream = HashedStream(fh)
    stream.write(b"srt was here")
    assert stream.hexdigest() == "cd7bd850d261f8fa39a41d0963b42dae5f303615db19ac79e5044586d0825b7b"


def test_asdf_stream_combined() -> None:
    fh = BytesIO()
    stream = Crc32Stream(fh)
    stream = CompressedStream(stream)
    stream = HashedStream(stream)

    stream.write(b"srt was here" * 100)
    stream.finalize()

    assert stream.hexdigest() == "ba40ab3ee826d6aa0f085dfccbb72d8feefa72548015c4456c1fd741d0266a94"
    assert fh.getvalue() == bytes.fromhex("789c2b2e2a51284f2c56c8482d4a2d1e658fb247d9a3ec41cc06004445c530665f35fc")


def test_asdf_scrape(asdf_writer: AsdfWriter) -> None:
    asdf_writer.add_bytes(b"\x00" * 0x1000, idx=0, base=0)
    asdf_writer.add_bytes(b"\x02" * 0x1000, idx=0, base=0x4000)
    asdf_writer.add_bytes(b"\x04" * 0x1000, idx=0, base=0x8000)
    asdf_writer.add_bytes(b"\x06" * 0x1000, idx=0, base=0x10000)
    asdf_writer.add_bytes(b"\xff" * 0x1000, idx=0, base=0x14000)

    # Don't close so we don't have a footer and block table
    asdf_writer._fh.seek(0)

    with pytest.raises(InvalidSnapshot):
        AsdfSnapshot(asdf_writer._fh)

    asdf_writer._fh.seek(0)
    reader = AsdfSnapshot(asdf_writer._fh, recover=True)
    stream = reader.open(0)

    assert [(run_start, run_size) for run_start, run_size, _, _ in stream.table] == [
        (0, 0x1000),
        (0x4000, 0x1000),
        (0x8000, 0x1000),
        (0x10000, 0x1000),
        (0x14000, 0x1000),
    ]

    assert stream.read() == b"".join(
        [
            (b"\x00" * 0x1000),
            (b"\xa5\xdf" * (0x3000 // 2)),
            (b"\x02" * 0x1000),
            (b"\xa5\xdf" * (0x3000 // 2)),
            (b"\x04" * 0x1000),
            (b"\xa5\xdf" * (0x7000 // 2)),
            (b"\x06" * 0x1000),
            (b"\xa5\xdf" * (0x3000 // 2)),
            (b"\xff" * 0x1000),
        ]
    )



from __future__ import annotations

import hashlib
from typing import BinaryIO

from dissect.evidence import ad1


def test_ad1(ad1_data: BinaryIO) -> None:
    a = ad1.AD1(ad1_data)

    assert a.header.magic == b"ADSEGMENTEDFILE\x00"
    assert a.root.name == b"E:\\AD1_test"
    assert len(a.root.children) == 2
    assert a.root.children[0].name == b"doc1.txt"
    assert a.root.children[0].open().read() == b"Inhoud document 1"


def test_ad1_long(ad1_data_long: BinaryIO) -> None:
    a = ad1.AD1(ad1_data_long)

    assert a.header.magic == b"ADSEGMENTEDFILE\x00"
    assert a.root.name == b"E:\\testdatamap 2 met spaties en een heel stuk langer"
    assert len(a.root.children) == 2

    entry = a.root.children[0]
    assert entry.name == b"een lange filenaam 1 met spaties.txt"
    assert entry.open().read() == (
        b"masdhdslkfjasdfjlksadjflkjsda;lfj\r\nasdflk\r\na;lsdkf\r\n"
        b";lasdklf;lkasd\r\n;lk\r\nfask;ldkf\r\n;lka\r\nsd;lkf\r\n"
        b"asdfasdaflkjsd;lkg;dfshglkdksfhg;ljsdflgjs;dlkkjg'qwjer'pgtoks\r\n"
        b"ddasd'dgkls'dfkjg\r\nsd'g;lkksd'f';gkjsd\r\n[fkgli'erjrg';ksd\r\n"
        b"'g'asldjg';askg\r\nkqe\r\n-["
    )
    md5sum = hashlib.md5(entry.open().read())
    assert md5sum.hexdigest().encode() == next(meta for meta in entry.meta if meta.type == ad1.MetaType.MD5).data


def test_ad1_compressed(ad1_data_compressed: BinaryIO) -> None:
    a = ad1.AD1(ad1_data_compressed)

    assert a.root.children[0].open().read() == b"Inhoud document 1"



from __future__ import annotations

import zlib
from typing import Any, BinaryIO

from dissect.cstruct import cstruct
from dissect.util.stream import AlignedStream, RelativeStream

ad1_def = """
enum EntryType : uint32 {
    File        = 0x0,
    Directory   = 0x5
};

enum MetaType : uint32 {
    FileClass           = 0x2,
    FileSize            = 0x3,
    PhysicalSize        = 0x4,
    StartCluster        = 0x6,
    DateAccessed        = 0x7,
    DateCreated         = 0x8,
    DateModified        = 0x9,
    Unknown_1           = 0xd,
    Unknown_2           = 0xe,
    ActualFile          = 0x1e,
    Unknown_3           = 0x1002,
    Unknown_4           = 0x1003,
    Unknown_5           = 0x1004,
    Unknown_6           = 0x1005,
    MD5                 = 0x5001,
    SHA1                = 0x5002,
    ClusterSize         = 0x9001,
    ClusterCount        = 0x9002,
    FreeClusterCount    = 0x9003,
    VolumeSerialNumber  = 0x9006
};

typedef struct {
    char        magic[16];
    uint32      unk1;
    uint32      unk2;
    uint32      unk3;
    uint32      unk4;
    uint16      unk5;
    uint16      version;
    uint32      unk6;
    uint64      logical_image_offset;
} SegmentedFileHeader;

typedef struct {
    char        magic[16];
    uint32      unk1;
    uint32      unk2;
    uint32      chunk_size;     // This is supposed to be uint64? But that doesn't seem right
    uint32      unk3;
    uint32      unk4;
    uint64      entry_offset;
    uint32      name_len;
    uint32      unk5;
    uint64      name_offset;
    uint64      unk6;
    uint64      unk7;
    uint64      unk8;
    uint64      unk9;
    char        name[name_len];
} LogicalImageHeader;

typedef struct {
    uint64      next;
    uint64      child;
    uint64      meta;
    uint64      unk1;
    uint64      size;
    EntryType   type;
    uint32      name_len;
    char        name[name_len];
    uint64      unk2;
    uint64      num_chunks;
    uint64      chunks[num_chunks];
} FileEntry;

typedef struct {
    uint64      next;
    uint32      category;
    MetaType    type;
    uint32      len;
    char        data[len];
} FileMeta;
"""
c_ad1 = cstruct().load(ad1_def)

EntryType = c_ad1.EntryType
MetaType = c_ad1.MetaType


class AD1:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.header = c_ad1.SegmentedFileHeader(fh)

        offset = self.header.logical_image_offset
        self.logical_image = LogicalImage(RelativeStream(fh, offset))
        self.root = self.logical_image

    def __getattr__(self, k: str) -> Any:
        if k in self.header.__class__.fields:
            return getattr(self.header, k)

        return super().__getattr__(k)

    def entry(self, path: str = "") -> LogicalImage | FileEntry:
        components = path.lstrip("/").split("/")
        current = self.root

        if components[0] == "":
            return current

        for c in components:
            for item in current.children:
                if item.name == c:
                    current = item

        if current.name == components[-1]:
            return current

        raise IOError(f"Path not found: {path}")

    def listdir(self, path: str) -> list[FileEntry]:
        return [e.name for e in self.entry(path).children]

    def get(self, path: str) -> LogicalImage | FileEntry:
        return self.entry(path)

    def open(self, path: str) -> FileObject:
        return self.entry(path).open()


class LogicalImage:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.header = c_ad1.LogicalImageHeader(fh)

        self.children = []
        offset = self.header.entry_offset
        while offset != 0:
            child = FileEntry(self, offset)
            offset = child.entry.next
            self.children.append(child)

    def __repr__(self) -> str:
        return f"<LogicalImage name={self.header.name}>"

    def __getattr__(self, k: str) -> Any:
        if k in self.header.__class__.fields:
            return getattr(self.header, k)

        return object.__getattribute__(self, k)


class FileEntry:
    def __init__(self, image: LogicalImage, offset: int):
        fh = image.fh
        fh.seek(offset)
        self.image = image
        self.offset = offset
        self.entry = c_ad1.FileEntry(fh)
        self.size = self.entry.size

        self.meta = []
        offset = self.entry.meta
        while offset != 0:
            meta = FileMeta(image, offset)
            offset = meta.entry.next
            self.meta.append(meta)

        self.children = []
        offset = self.entry.child
        while offset != 0:
            child = FileEntry(image, offset)
            offset = child.entry.next
            self.children.append(child)

    def __repr__(self) -> str:
        file_type = "Unknown type"
        if self.is_file():
            file_type = "File"
        elif self.is_dir():
            file_type = "Directory"
        return f"<{file_type} name={self.entry.name}>"

    def __getattr__(self, k: str) -> Any:
        if k in self.entry.__class__.fields:
            return getattr(self.entry, k)

        return object.__getattribute__(self, k)

    def open(self) -> FileObject:
        return FileObject(self)

    def is_file(self) -> bool:
        return self.entry.type == EntryType.File

    def is_dir(self) -> bool:
        return self.entry.type == EntryType.Directory


class FileMeta:
    def __init__(self, image: LogicalImage, offset: int):
        fh = image.fh
        fh.seek(offset)
        self.image = image
        self.offset = offset
        self.entry = c_ad1.FileMeta(fh)

    def __repr__(self) -> str:
        return f"<Meta category={self.entry.category} type={self.entry.type} data={self.entry.data}>"

    def __getattr__(self, k: str) -> Any:
        if k in self.entry.__class__.fields:
            return getattr(self.entry, k)

        return object.__getattribute__(self, k)


class FileObject(AlignedStream):
    def __init__(self, entry: FileEntry):
        self.entry = entry
        super().__init__(self.entry.size, self.entry.image.chunk_size)

    def _read(self, offset: int, length: int) -> bytes:
        r = []
        fh = self.entry.image.fh
        chunk_size = self.entry.image.chunk_size

        chunk = offset // chunk_size
        chunk_count = (length + chunk_size - 1) // chunk_size

        chunk_offsets = self.entry.entry.chunks[chunk : chunk + chunk_count + 1]
        if len(chunk_offsets) != chunk_count + 1:
            chunk_offsets.append(self.entry.entry.meta)

        fh.seek(chunk_offsets[0])
        buf = fh.read(chunk_offsets[-1] - chunk_offsets[0])

        prev_offset = chunk_offsets[0]
        for offset in chunk_offsets[1:]:
            chunk_size = offset - prev_offset
            r.append(zlib.decompress(buf[:chunk_size]))
            buf = buf[chunk_size:]
            prev_offset = offset

        return b"".join(r)



from __future__ import annotations

import argparse
import datetime
import hashlib
import io
import shutil
import stat
import sys
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.evidence.asdf import asdf

if TYPE_CHECKING:
    from collections.abc import Iterator


def iterfileobj(src: BinaryIO) -> Iterator[bytes]:
    buf = src.read(io.DEFAULT_BUFFER_SIZE)
    while buf:
        yield buf
        buf = src.read(io.DEFAULT_BUFFER_SIZE)


def hashfileobj(src: BinaryIO, alg: str = "sha256") -> str:
    ctx = hashlib.new(alg)
    for buf in iterfileobj(src):
        ctx.update(buf)
    return ctx.hexdigest()


def handle_ls(snapshot: asdf.AsdfSnapshot, args: argparse.Namespace) -> None:
    for member in snapshot.metadata.members():
        mtime = datetime.datetime.fromtimestamp(member.mtime, datetime.timezone.utc).isoformat()

        print(f"{stat.filemode(member.mode)} {member.size:9d} {mtime} {member.name}")


def handle_cat(snapshot: asdf.AsdfSnapshot, args: argparse.Namespace) -> None:
    try:
        fh = snapshot.metadata.open(args.name)
    except Exception:
        print(f"file doesn't exist: {args.name}")

    stdout = sys.stdout.buffer if hasattr(sys.stdout, "buffer") else sys.stdout
    try:
        shutil.copyfileobj(fh, stdout)
    except BrokenPipeError:
        pass


def handle_hash(snapshot: asdf.AsdfSnapshot, args: argparse.Namespace) -> None:
    for name in snapshot.metadata.names():
        print(f"{hashfileobj(snapshot.metadata.open(name))}  {name}")


def main() -> int:
    parser = argparse.ArgumentParser(description="Utility to work with ASDF metadata")
    parser.add_argument("file", metavar="ASDF", help="ASDF file to verify")
    subparsers = parser.add_subparsers(help="metadata command")
    parser_ls = subparsers.add_parser("ls", help="list metadata contents")
    parser_ls.set_defaults(handler=handle_ls)

    parser_cat = subparsers.add_parser("cat", help="cat metadata contents")
    parser_cat.add_argument("name", help="file to cat")
    parser_cat.set_defaults(handler=handle_cat)

    parser_hash = subparsers.add_parser("hash", help="hash metadata contents")
    parser_hash.set_defaults(handler=handle_hash)
    args = parser.parse_args()

    if not hasattr(args, "handler"):
        parser.print_help()
        return 1

    with Path(args.file).open("rb") as fh:
        snapshot = asdf.AsdfSnapshot(fh)
        args.handler(snapshot, args)

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

import logging
import os
import zlib
from bisect import bisect_right
from functools import lru_cache
from pathlib import Path
from typing import BinaryIO

from dissect.cstruct import cstruct
from dissect.util.stream import AlignedStream

from dissect.evidence.exceptions import EWFError

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_EWF", "CRITICAL"))

ewf_def = """
enum MediaType : uint8 {
    Removable   = 0x00,
    Fixed       = 0x01,
    Optical     = 0x03,
    Logical     = 0x0e,
    RAM         = 0x10
};

enum MediaFlags : uint8 {
    Image       = 0x01,
    Physical    = 0x02,
    Fastbloc    = 0x04,
    Tablaeu     = 0x08
};

enum CompressionLevel : uint8 {
    None        = 0x00,
    Good        = 0x01,
    Best        = 0x02
};

typedef struct {
    char        signature[8];
    uint8       fields_start;
    uint16      segment_number;
    uint16      fields_end;
} EWFHeader;

typedef struct {
    char    type[16];
    uint64  next;
    uint64  size;
    uint8   pad[40];
    uint32  checksum;
} EWFSectionDescriptor;

typedef struct {
    uint32  reserved_1;
    uint32  chunk_count;
    uint32  sector_count;
    uint32  sector_size;
    uint32  total_sector_count;
    uint8   reserved[20];
    uint8   pad[45];
    char    signature[5];
    uint32  checksum;
} EWFVolumeSectionSpec;

typedef struct {
    MediaType           media_type;
    uint8               reserved_1[3];
    uint32              chunk_count;
    uint32              sector_count;
    uint32              sector_size;
    uint64              total_sector_count;
    uint32              num_cylinders;
    uint32              num_heads;
    uint32              num_sectors;
    uint8               media_flags;
    uint8               unknown_1[3];
    uint32              palm_start_sector;
    uint32              unknown_2;
    uint32              smart_start_sector;
    CompressionLevel    compression_level;
    uint8               unknown_3[3];
    uint32              error_granularity;
    uint32              unknown_4;
    uint8               uuid[16];
    uint8               pad[963];
    char                signature[5];
    uint32              checksum;
} EWFVolumeSection;

typedef struct {
    uint32  num_entries;
    uint32  _;
    uint64  base_offset;
    uint32  _;
    uint32  checksum;
    uint32  entries[num_entries];
} EWFTableSection;
"""

c_ewf = cstruct().load(ewf_def)

MAX_OPEN_SEGMENTS = 128


def find_files(path: str | Path) -> list[Path]:
    """Find all related EWF files from the given path."""
    if not isinstance(path, Path):
        path = Path(path)

    path = path.resolve()
    ext = path.suffix

    if ext[1].upper() not in "ELS":
        raise EWFError(f"Invalid EWF file: {path}")

    ewfglob = f"[{ext[1]}-{'Z' if ext[1].isupper() else 'z'}]" if len(ext) == 4 else f"{ext[1]}[x-z]"

    return sorted(path.parent.glob(f"{path.stem}.{ewfglob}[0-9A-Za-z][0-9A-Za-z]"))


class EWF:
    """Expert Witness Disk Image Format."""

    def __init__(self, fh: BinaryIO | list[BinaryIO]):
        fhs = [fh] if not isinstance(fh, list) else fh

        self.fh = fhs
        self.header: HeaderSection = None
        self.volume: VolumeSection = None
        self._segments: dict[str, Segment] = {}
        self._segment_offsets = []
        self._segment_lru = []

        segment_offset = 0

        for i in range(len(fhs)):
            try:
                segment = self.open_segment(i)
            except Exception:
                log.exception("Failed to parse as EWF file: %s", fh)
                continue

            if segment.header and not self.header:
                self.header = segment.header

            if segment.volume and not self.volume:
                self.volume = segment.volume

            if segment_offset != 0:
                self._segment_offsets.append(segment_offset)

            segment.offset = segment_offset * self.volume.sector_size
            segment.sector_offset = segment_offset
            segment_offset += segment.sector_count

        if not self.header or not self.volume:
            raise EWFError(f"Failed to load EWF: {fh}")

        self.chunk_size = self.volume.sector_count * self.volume.sector_size

        max_size = self.volume.chunk_count * self.volume.sector_count * self.volume.sector_size
        last_table = self.open_segment(len(self.fh) - 1).tables[-1]
        last_chunk_size = len(last_table.read_chunk(last_table.num_entries - 1))

        self.size = max_size - (self.chunk_size - last_chunk_size)

    def open_segment(self, idx: int) -> Segment:
        # Poor mans LRU
        if idx in self._segments:
            self._segment_lru.remove(idx)
            self._segment_lru.append(idx)
            return self._segments[idx]

        if len(self._segment_lru) >= MAX_OPEN_SEGMENTS:
            oldest_idx = self._segment_lru.pop(0)
            oldest_segment = self._segments.pop(oldest_idx)

            # Don't close it if we received it as a file-like object
            if not hasattr(self.fh[oldest_idx], "read"):
                oldest_segment.fh.close()

            del oldest_segment

        fh = self.fh[idx]
        if not hasattr(fh, "read"):
            fh = fh.open("rb") if isinstance(fh, Path) else Path(fh).open("rb")  # noqa: SIM115

        segment = Segment(self, fh)
        if self.volume and 0 < idx <= len(self._segment_offsets):
            # We already have a known segment offset for this segment, so set it back
            segment_offset = self._segment_offsets[idx - 1]
            segment.offset = segment_offset * self.volume.sector_size
            segment.sector_offset = segment_offset
        else:
            # Otherwise we're in the initialization loop (or we're idx == 0)
            segment.offset = 0
            segment.sector_offset = 0

        self._segments[idx] = segment
        self._segment_lru.append(idx)

        return segment

    def open(self) -> BinaryIO:
        return EWFStream(self)


class EWFStream(AlignedStream):
    def __init__(self, ewf: EWF):
        self.ewf = ewf
        self.sector_size = self.ewf.volume.sector_size
        super().__init__(ewf.size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        sector_offset = offset // self.sector_size
        sector_count = (length + self.sector_size - 1) // self.sector_size

        segment_idx = bisect_right(self.ewf._segment_offsets, sector_offset)

        while sector_count > 0:
            if segment_idx > len(self.ewf._segment_offsets):
                raise EWFError(f"Missing EWF file for segment index: {segment_idx}")

            segment = self.ewf.open_segment(segment_idx)

            segment_remaining_sectors = segment.sector_count - (sector_offset - segment.sector_offset)
            segment_sectors = min(segment_remaining_sectors, sector_count)

            result.append(segment.read_sectors(sector_offset, segment_sectors))
            sector_offset += segment_sectors
            sector_count -= segment_sectors

            segment_idx += 1

        return b"".join(result)


class Segment:
    def __init__(self, ewf: EWF, fh: BinaryIO):
        self.ewf = ewf
        self.fh = fh

        fh.seek(0)
        self.ewfheader = c_ewf.EWFHeader(fh)
        self.header = ewf.header
        self.volume = ewf.volume

        if self.ewfheader.signature not in (b"EVF\x09\x0d\x0a\xff\x00", b"LVF\x09\x0d\x0a\xff\x00"):
            raise EWFError(f"Invalid signature, got {self.ewfheader.signature!r}")

        self.sections: list[SectionDescriptor] = []
        self.tables: list[TableSection] = []
        self.table_offsets = []

        offset = 0
        sector_offset = 0

        while True:
            section = SectionDescriptor(fh)
            self.sections.append(section)

            if section.type in (b"header", b"header2") and not self.header:
                self.header = HeaderSection(self, section)

            if section.type in (b"disk", b"volume") and not self.volume:
                self.volume = VolumeSection(self, section)

            if section.type == b"table":
                table = TableSection(self, section)

                if sector_offset != 0:
                    self.table_offsets.append(sector_offset)

                table.offset = sector_offset * self.volume.sector_size
                table.sector_offset = sector_offset
                sector_offset += table.sector_count

                self.tables.append(table)

            if section.next == offset or section.type == b"done":
                break

            offset = section.next
            fh.seek(offset)

        self.chunk_count = sum([t.num_entries for t in self.tables])
        self.sector_count = self.chunk_count * self.volume.sector_count
        self.size = self.chunk_count * self.volume.sector_count * self.volume.sector_size
        self.sector_offset = None  # Set later
        self.offset = None  # Set later

    def read_sectors(self, sector: int, count: int) -> bytes:
        log.debug("Segment::read_sectors(0x%x, 0x%x)", sector, count)
        segment_sector = sector - self.sector_offset
        r = []

        table_idx = bisect_right(self.table_offsets, segment_sector)
        while count > 0:
            table = self.tables[table_idx]

            table_remaining_sectors = table.sector_count - (segment_sector - table.sector_offset)
            table_sectors = min(table_remaining_sectors, count)

            r.append(table.read_sectors(segment_sector, table_sectors))
            segment_sector += table_sectors
            count -= table_sectors

            table_idx += 1

        return b"".join(r)


class HeaderSection:
    def __init__(self, segment: Segment, section: SectionDescriptor):
        self.segment = segment
        self.section = section

        fh = segment.fh
        fh.seek(section.data_offset)
        self.data = zlib.decompress(fh.read(section.size))

        if self.data[0] in (b"\xff", b"\xfe"):
            self.data = self.data.decode("utf16")

    def __repr__(self) -> str:
        return f"<HeaderSection categories={int(self.data[0])}>"


class VolumeSection:
    def __init__(self, segment: Segment, section: SectionDescriptor):
        self.segment = segment
        self.section = section

        fh = segment.fh
        fh.seek(section.data_offset)
        data = c_ewf.EWFVolumeSection(fh) if section.size == 1052 else c_ewf.EWFVolumeSectionSpec(fh)

        self.volume = data
        self.chunk_count = data.chunk_count
        self.sector_count = data.sector_count
        self.sector_size = data.sector_size


class TableSection:
    def __init__(self, segment: Segment, section: SectionDescriptor):
        self.segment = segment
        self.section = section

        fh = segment.fh
        fh.seek(section.data_offset)

        self.header = c_ewf.EWFTableSection(fh)
        self.num_entries = self.header.num_entries
        self.base_offset = self.header.base_offset
        self.entries = self.header.entries

        self.sector_count = self.num_entries * self.segment.volume.sector_count
        self.size = self.sector_count * self.segment.volume.sector_size
        self.sector_offset = None  # Set later
        self.offset = None  # Set later

        self.read_chunk = lru_cache(1024)(self.read_chunk)

    def read_chunk(self, chunk: int) -> bytes:
        log.debug("TableSection::read_chunk(0x%x)", chunk)

        chunk_entry = self.entries[chunk]
        chunk_offset = self.base_offset + (chunk_entry & 0x7FFFFFFF)
        compressed = chunk_entry >> 31 == 1

        # EWF sucks
        # We don't know the chunk size, so try to determine it using the offset of the next chunk
        # When it's the last chunk in the table though, this becomes trickier.
        # We have to check if the chunk data is preceding the table, or if it's contained within the table section
        # Then we can calculate the chunk size using these offsets
        if chunk + 1 == self.num_entries:
            # The chunk data is stored before the table section
            if chunk_offset < self.section.offset:
                chunk_size = self.section.offset - chunk_offset
            # The chunk data is stored within the table section
            elif chunk_offset < self.section.offset + self.section.size:
                chunk_size = self.section.offset + self.section.size - chunk_offset
            else:
                raise EWFError("Unknown size of last chunk")
        else:
            chunk_size = self.base_offset + (self.entries[chunk + 1] & 0x7FFFFFFF) - chunk_offset

        # Non compressed chunks have a 4 byte checksum
        if not compressed:
            chunk_size -= 4

        self.segment.fh.seek(chunk_offset)
        buf = self.segment.fh.read(chunk_size)

        if compressed:
            buf = zlib.decompress(buf)

        return buf

    def read_sectors(self, sector: int, count: int) -> bytes:
        log.debug("TableSection::read_sectors(0x%x, 0x%x)", sector, count)
        result = []

        chunk_sector_count = self.segment.volume.sector_count
        sector_size = self.segment.volume.sector_size

        table_sector = sector - self.sector_offset
        table_chunk = table_sector // chunk_sector_count

        while count > 0:
            table_sector_offset = table_sector % chunk_sector_count
            chunk_remaining_sectors = chunk_sector_count - table_sector_offset
            table_sectors = min(chunk_remaining_sectors, count)

            chunk_pos = table_sector_offset * sector_size
            chunk_end = chunk_pos + (table_sectors * sector_size)

            buf = self.read_chunk(table_chunk)
            if chunk_pos != 0 or table_sectors != chunk_sector_count:
                buf = buf[chunk_pos:chunk_end]
            result.append(buf)

            count -= table_sectors
            table_sector += table_sectors
            table_chunk += 1

        return b"".join(result)


class SectionDescriptor:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        self.offset = fh.tell()
        descriptor = c_ewf.EWFSectionDescriptor(fh)
        self.type = descriptor.type.rstrip(b"\x00")
        self.next = descriptor.next
        self.size = descriptor.size - len(c_ewf.EWFSectionDescriptor)
        self.checksum = descriptor.checksum
        self.data_offset = fh.tell()

    def __repr__(self) -> str:
        return (
            f"<SectionDescriptor "
            f"type={self.type} size={self.size:#x} offset={self.offset:#x} checksum={self.checksum:#x}>"
        )



from __future__ import annotations

import argparse
import io
import shutil
import sys
from contextlib import nullcontext
from pathlib import Path

from dissect.evidence.asdf import asdf
from dissect.evidence.asdf.streams import HashedStream


def main() -> int:
    parser = argparse.ArgumentParser(description="Utility to repair block tables of ASDF files")
    parser.add_argument("file", metavar="ASDF", help="ASDF file to repair")
    parser.add_argument("-w", "--writer", default="-", help="file to write to, default is stdout")
    args = parser.parse_args()

    with Path(args.file).open("rb") as fh:
        header = asdf.c_asdf.header(fh)
        if header.magic != asdf.FILE_MAGIC:
            parser.exit(1, "invalid header magic, is this an ASDF file?")

        fh.seek(-len(asdf.c_asdf.footer), io.SEEK_END)
        footer = asdf.c_asdf.footer(fh)
        if footer.magic == asdf.FOOTER_MAGIC:
            parser.exit(1, "ASDF file already has a valid footer, no repair necessary")

        fh.seek(len(asdf.c_asdf.header))

        table = []
        prev_entry = None
        for block, file_offset in asdf.scrape_blocks(fh):
            entry = asdf.c_asdf.table_entry(
                flags=block.flags,
                idx=block.idx,
                offset=block.offset,
                size=block.size,
                file_offset=file_offset,
                file_size=0,  # Filled in later
            )
            table.append(entry)

            if prev_entry:
                prev_entry.file_size = (file_offset - prev_entry.file_offset) - len(asdf.c_asdf.block)

            prev_entry = entry

        if prev_entry:
            prev_entry.file_size = (fh.seek(0, io.SEEK_END) - prev_entry.file_offset) - len(asdf.c_asdf.block)

        print(f"[*] Recovered {len(table)} blocks", file=sys.stderr)

        ctx = nullcontext(getattr(sys.stdout, "buffer", sys.stdout)) if args.writer == "-" else None
        with ctx or Path(args.writer).open("wb") as fhout:
            try:
                print(f"[*] Writing to {args.writer}", file=sys.stderr)
                # Copy the original file
                fh.seek(0)
                fhout = HashedStream(fhout)
                shutil.copyfileobj(fh, fhout)

                print("[*] Writing block table", file=sys.stderr)
                # Write the block table
                table_offset = fh.tell()  # fhout might be stdout which we can't tell() on
                for entry in table:
                    entry.write(fhout)

                print("[*] Writing footer", file=sys.stderr)
                # Write the footer
                asdf.c_asdf.footer(
                    magic=asdf.FOOTER_MAGIC,
                    table_offset=table_offset,
                    sha256=fhout.digest(),
                ).write(fhout)

                print("[*] All done, use asdf-verify to check file", file=sys.stderr)
            except BrokenPipeError:
                pass

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass
